import asyncio
import os
from types import ModuleType
from typing import Any, Optional
from sqlalchemy import Engine
from sqlmodel import create_engine as alchemy_create_engine
from sqlalchemy.ext.asyncio import (
    create_async_engine as alchemy_create_engine_async,
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
)
from sqlalchemy.orm import Session
from contextvars import ContextVar

engines_sync: ContextVar[dict[str, Engine]] = ContextVar("engines_sync")
engines_async: ContextVar[dict[str, AsyncEngine]] = ContextVar("engines_async")


def create_engine_sync(module, **kwargs):
    # get the name
    db_name = module.__name__.split(".")[-1].lower()
    dialect = module.__name__.split(".")[-2].lower()
    db_path = (
        os.path.dirname(module.__file__).replace(os.getcwd(), "").replace("\\", "/")
    )
    # connect to database
    engine = alchemy_create_engine(f"{dialect}://{db_path}/{db_name}.db", **kwargs)
    return engine


def create_engine_async(module, driver, **kwargs) -> AsyncEngine:  # type: ignore
    db_name = module.__name__.split(".")[-1].lower()
    dialect = module.__name__.split(".")[-2].lower()
    db_path = (
        os.path.dirname(module.__file__).replace(os.getcwd(), "").replace("\\", "/")
    )
    if dialect == "sqlite":
        print("sqlite")
        return alchemy_create_engine_async(
            f"{dialect}+{driver}://{db_path}/{db_name}.db", **kwargs
        )


def init_database_sync(engine, module):
    # initialize database
    module.SQLModel.metadata.create_all(engine)


async def init_database_async(engine: AsyncEngine, module: Any) -> None:
    """
    Initialize the database asynchronously.

    Args:
        engine (Engine): The SQLAlchemy engine instance.
        module (Any): The module containing the SQLModel metadata.
    """
    async with engine.begin() as connection:
        await connection.run_sync(module.SQLModel.metadata.create_all)


def createSyncEngine(module: ModuleType, **kwargs):
    """
    create new engine and add it to context var
    """

    engine = create_engine_sync(module, **kwargs)

    engines_dict: dict[str, Engine] = engines_sync.get({})

    engines_dict[module.__name__.split(".")[-1].lower()] = engine

    engines_sync.set(engines_dict)
    return


def createAsyncEngine(module: ModuleType, driver: str, **kwargs):
    """
    create new engine and add it to context var
    """

    engine = create_engine_async(module, driver, **kwargs)

    engines_dict: dict[str, AsyncEngine] = engines_async.get({})

    engines_dict[module.__name__.split(".")[-1].lower()] = engine

    engines_async.set(engines_dict)
    return


def getSyncEngine(database: ModuleType | str) -> Engine:
    if isinstance(database, str):
        engine: Engine = engines_sync.get()[database.lower()]
    elif isinstance(database, ModuleType):
        engine: Engine = engines_sync.get()[database.__name__.split(".")[-1].lower()]
    return engine


def getAsyncEngine(database: ModuleType | str) -> AsyncEngine:
    if isinstance(database, str):
        engine: AsyncEngine = engines_async.get()[database.lower()]
    elif isinstance(database, ModuleType):
        engine: AsyncEngine = engines_async.get()[
            database.__name__.split(".")[-1].lower()
        ]
    return engine


async def asyncDbSession(database: ModuleType | str) -> AsyncSession:
    # get the engine
    if isinstance(database, str):
        engine: AsyncEngine = getAsyncEngine(database)
    elif isinstance(database, ModuleType):
        engine: AsyncEngine = getAsyncEngine(database.__name__.split(".")[-1].lower())
    return AsyncSession(engine)


def syncDbSession(database: ModuleType | str) -> Session:
    # get the engine
    if isinstance(database, str):
        engine: Engine = getSyncEngine(database)
    elif isinstance(database, ModuleType):
        engine: Engine = getSyncEngine(database.__name__.split(".")[-1].lower())

    return Session(engine)


"""
if __name__ == "__main__":
    new_engine_sync: Engine = create_engine_sync(TodoDatabase)
    new_engine_async: AsyncEngine = create_engine_async(TodoDatabase, "aiosqlite")
    asyncio.run(init_database_async(new_engine_async, TodoDatabase))
"""



from contextlib import asynccontextmanager
from types import ModuleType
import importlib.util
import os
from fastapi import APIRouter, FastAPI
from typing import Any, AsyncGenerator, Callable, Optional

from typing import List

FILE_CONVENTIONS: List[str] = [
    "socket.py",
    "config.py",
    "startup.py",
    "shutdown.py",
    "get.py",
    "post.py",
    "put.py",
    "delete.py",
    "patch.py",
    "options.py",
    "head.py",
    "license.md",
    "connection.py",
]

FASTAPI_PARAMETERS: List[str] = [
    "debug",
    "routes",
    "title",
    "summary",
    "description",
    "version",
    "openapi_url",
    "openapi_tags",
    "servers",
    "dependencies",
    "default_response_class",
    "redirect_slashes",
    "docs_url",
    "redoc_url",
    "swagger_ui_oauth2_redirect_url",
    "swagger_ui_init_oauth",
    "middleware",
    "exception_handlers",
    "on_startup",
    "on_shutdown",
    "lifespan",
    "terms_of_service",
    "contact",
    "license_info",
    "openapi_prefix",
    "root_path",
    "root_path_in_servers",
    "responses",
    "callbacks",
    "webhooks",
    "deprecated",
    "include_in_schema",
    "swagger_ui_parameters",
    "generate_unique_id_function",
    "separate_input_output_schemas",
]

API_ROUTER_PARAMETERS: List[str] = [
    "prefix",
    "tags",
    "dependencies",
    "default_response_class",
    "responses",
    "callbacks",
    "routes",
    "redirect_slashes",
    "default",
    "dependency_overrides_provider",
    "route_class",
    "deprecated",
    "include_in_schema",
    "generate_unique_id_function",
]

API_HTTP_VERBS: List[str] = ["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "HEAD"]

API_HTTP_PARAMETERS: List[str] = [
    "response_model",
    "status_code",
    "tags",
    "dependencies",
    "summary",
    "description",
    "response_description",
    "responses",
    "deprecated",
    "operation_id",
    "response_model_include",
    "response_model_exclude",
    "response_model_by_alias",
    "response_model_exclude_unset",
    "response_model_exclude_defaults",
    "response_model_exclude_none",
    "include_in_schema",
    "response_class",
    "name",
    "callbacks",
    "openapi_extra",
    "generate_unique_id_function",
]


class Pypox:
    def __init__(self, directory: str) -> None:
        """
        Initializes an instance of the class.

        Parameters:
            directory (str): The directory path.

        Returns:
            None
        """
        self.directory: str = directory
        self.main_api: Optional[FastAPI] = None  # type: ignore
        self.python_modules: dict[str, List[ModuleType]] = {}
        self.get_modules()

    def get_modules(self) -> "Pypox":
        """
        Get the list of modules in the specified directory.

        Returns:
            Pypox: The current instance of Pypox.
        """
        for root, dirs, files in os.walk(f"{self.directory}"):
            folder_modules: List[ModuleType] = []
            for file in files:
                if file not in FILE_CONVENTIONS:
                    continue
                if ".py" in file:
                    module_name: str = os.path.splitext(file)[0]
                    module_path: str = os.path.join(root, file)
                    spec: ModuleType = importlib.util.spec_from_file_location(module_name, module_path)  # type: ignore
                    module: ModuleType = importlib.util.module_from_spec(spec)  # type: ignore
                    spec.loader.exec_module(module)  # type: ignore
                    folder_modules.append(module)

                    self.python_modules[
                        root.replace("[", "{")
                        .replace("]", "}")
                        .replace("\\routes\\", "/")
                    ] = folder_modules
        return self

    def __create_lifespan(self, modules: List[ModuleType]) -> Callable | None:
        """
        Creates a lifespan context manager function based on the provided modules.

        Parameters:
            modules (List[ModuleType]): The list of modules to check for the presence of "startup" and "shutdown" functions.

        Returns:
            Callable | None: The lifespan context manager function if both "startup" and "shutdown" functions are present in the modules, otherwise None.
        """
        if "startup" not in [x.__name__ for x in modules] or "shutdown" not in [
            x.__name__ for x in modules
        ]:
            return None

        @asynccontextmanager
        async def lifespan(app: FastAPI) -> AsyncGenerator[None, Any]:
            await [x.__call__ for x in modules if x.__name__ == "startup"][0](app)
            yield
            await [x.__call__ for x in modules if x.__name__ == "shutdown"][0](app)

        return lifespan

    def __create_config(self, modules: List[ModuleType], type: str) -> dict[str, Any]:
        config: dict[str, Any] = {}
        """
        Generates a configuration dictionary based on the given modules and type.

        Args:
            modules (List[Type[ModuleType]]): The list of modules to generate the configuration from.
            type (str): The type of configuration to generate. Valid values are "FastAPI" and "API_ROUTER".

        Returns:
            Dict[str, Any]: The generated configuration dictionary.
        """
        parameters: List[str] = (
            FASTAPI_PARAMETERS if type == "FastAPI" else API_ROUTER_PARAMETERS
        )

        for module in modules:
            for parameter in parameters:
                if hasattr(module, parameter):
                    config[parameter] = getattr(module, parameter)

        return config

    def __get_endpoints(
        self, modules: list[ModuleType]
    ) -> tuple[list[Callable], list[dict[str, Any]]]:
        """
        Generates a list of endpoints and their corresponding configurations based on a list of modules.

        Args:
            modules (list[ModuleType]): A list of modules containing endpoint information.

        Returns:
            tuple[list[Callable], list[dict[str, Any]]]: A tuple containing two lists:
                - endpoints: A list of endpoint functions.
                - configs: A list of endpoint configurations.

        Raises:
            AssertionError: If an endpoint is not found in a module.
        """
        endpoints: list[Callable] = []
        configs: list[dict[str, Any]] = []
        for module in modules:
            if (
                module.__name__.upper() not in API_HTTP_VERBS
                and module.__name__ != "socket"
            ):
                continue

            assert hasattr(module, "endpoint"), "endpoint not found in module"
            endpoints.append(getattr(module, "endpoint"))
            api_config: dict[str, Any] = {
                key: getattr(module, key)
                for key in API_HTTP_PARAMETERS
                if hasattr(module, key)
            }
            api_config["methods"] = [module.__name__.upper()]
            configs.append(api_config)
        return endpoints, configs

    def __call__(self, *args: Any, **kwds: Any) -> FastAPI:
        """
        Execute the function when the class instance is called.

        Args:
            *args: Variable length arguments.
            **kwds: Keyword arguments.

        Returns:
            Any: The return value of the function.
        """
        for root, modules in self.python_modules.items():
            endpoints, configs = self.__get_endpoints(modules)
            if root == self.directory:
                lifespan: Callable[[], None] | None = self.__create_lifespan(modules)
                config: dict[str, Any] = self.__create_config(modules, "FastAPI")
                self.main_api: FastAPI = FastAPI(**config, lifespan=lifespan)

            if self.main_api and root == self.directory:
                for endpoint, endpoint_config in zip(endpoints, configs):
                    if endpoint_config["methods"][0] == "SOCKET":
                        self.main_api.add_api_websocket_route(
                            path=root.replace(self.directory, "").replace("\\", "/")
                            + "/",
                            endpoint=endpoint,
                            **{
                                key: val
                                for key, val in endpoint_config.items()
                                if key != "methods"
                            },
                        )
                        continue
                    self.main_api.add_api_route(
                        path=root.replace(self.directory, "").replace("\\", "/") + "/",
                        name=f"{endpoint_config['methods'][0]}_Endpoint",
                        endpoint=endpoint,
                        **endpoint_config,
                    )
            elif self.main_api and root != self.directory:
                config: dict[str, Any] = self.__create_config(modules, "APIRouter")
                router: APIRouter = APIRouter(**config)
                for endpoint, endpoint_config in zip(endpoints, configs):
                    if endpoint_config["methods"][0] == "SOCKET":
                        self.main_api.add_api_websocket_route(
                            path=root.replace(self.directory, "").replace("\\", "/")
                            + "/",
                            endpoint=endpoint,
                            **{
                                key: val
                                for key, val in endpoint_config.items()
                                if key != "methods"
                            },
                        )
                        continue
                    router.add_api_route(
                        path=root.replace(self.directory, "").replace("\\", "/") + "/",
                        name=f"{endpoint_config['methods'][0]}_Endpoint",
                        endpoint=endpoint,
                        **endpoint_config,
                    )
                self.main_api.include_router(router)

        return self.main_api



import asyncio
from typing import AsyncGenerator
from click import Context
from fastapi import FastAPI
from sqlalchemy.ext.asyncio import AsyncSession
from test.app.database.SQLITE import TodoDatabase
from pypox.database import (
    createAsyncEngine,
    getAsyncEngine,
    init_database_async,
)

createAsyncEngine(TodoDatabase, "aiosqlite")


async def __call__(app: FastAPI) -> None:
    # create async engine
    print("starting system")
    await init_database_async(getAsyncEngine(TodoDatabase), TodoDatabase)



import os
import random
from httpx import AsyncClient
from test.app.main import app
import pytest
import faker


# test all routes

test_user: dict[str, dict] = {}
SIZE = 10


# test register rout
@pytest.mark.asyncio
async def test_register():
    async with AsyncClient(
        app=app,
        base_url="http://localhost:8000",
    ) as ac:
        fake_user_gen = faker.Faker()
        for _ in range(SIZE):
            fake_user = fake_user_gen.simple_profile()
            test_user[fake_user["username"]] = {
                "username": fake_user["username"],
                "password": f"{random.randint(0, 100000000)}",
                "name": fake_user["name"],
                "email": fake_user["mail"],
                "phone": f"{random.randint(0, 100000000000)}",
            }

        for key in test_user:
            response = await ac.post(
                "/auth/register/",
                json=test_user[key],
            )
            assert response.status_code == 200 or response.status_code == 409


@pytest.mark.asyncio
async def test_login():
    async with AsyncClient(app=app, base_url="http://localhost:8000") as ac:
        for key in test_user:
            response = await ac.post(
                "/auth/login/",
                json={
                    "username": test_user[key]["username"],
                    "password": test_user[key]["password"],
                },
            )
            test_user[key]["id"] = response.text
            assert response.status_code == 200


@pytest.mark.asyncio
async def test_postTodo():
    async with AsyncClient(app=app, base_url="http://localhost:8000") as ac:
        fake_user_gen = faker.Faker()
        for key in test_user:
            response = await ac.post(
                f"/todo/?user_id={test_user[key]['id']}",
                json={
                    "title": fake_user_gen.text(max_nb_chars=20),
                    "description": fake_user_gen.paragraph(nb_sentences=3),
                    "completed": False,
                },
            )
            assert response.status_code == 200


@pytest.mark.asyncio
async def test_getTodo():
    async with AsyncClient(app=app, base_url="http://localhost:8000") as ac:
        for key in test_user:
            response = await ac.get(
                f"/todo/?user_id={test_user[key]['id']}",
            )
            test_user[key]["todo"] = response.json()
            assert response.status_code == 200


@pytest.mark.asyncio
async def test_updateTodo():
    async with AsyncClient(app=app, base_url="http://localhost:8000") as ac:
        for key in test_user:
            for todo in test_user[key]["todo"]:
                response = await ac.put(
                    f"/todo/{todo.get('id')}/?user_id={test_user[key]['id']}",
                    json={
                        "title": todo.get("title"),
                        "description": todo.get("description"),
                        "completed": True,
                    },
                )
                assert response.status_code == 200


@pytest.mark.asyncio
async def test_deleteTodo():
    async with AsyncClient(app=app, base_url="http://localhost:8000") as ac:
        for key in test_user:
            for todo in test_user[key]["todo"]:
                response = await ac.delete(
                    f"/todo/{todo.get('id')}/?user_id={test_user[key]['id']}",
                )
                assert response.status_code == 200



from setuptools import setup, find_packages

setup(
    name="pypox",
    version="1.3.1",
    author="Karl Robeck Alferez",
    author_email="karlalferezfx@gmail.com",
    description="",
    packages=find_packages(),
    install_requires=["uvicorn[standard]", "fastapi"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11.0",
)
